/*
 * Part of Intel(R) Manageability Engine Interface Linux driver
 *
 * Copyright (c) 2003 - 2008 Intel Corp.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */

#ifndef _HECI_DATA_STRUCTURES_H_
#define _HECI_DATA_STRUCTURES_H_

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/moduleparam.h>
#include "kcompat.h"

/*
 * error code definition
 */
#define     ESLOTS_OVERFLOW              1
#define     ECORRUPTED_MESSAGE_HEADER    1000
#define     EMESSAGE_NOT_COMPLETE        1001
#define     EMESSAGE_WILL_SPLIT          1002

#define     HECI_FC_MESSAGE_RESERVED_LENGTH           5

/*
 * Maximum transmission unit (MTU) of heci messages
 */
#define IAMTHIF_MTU 4160


/*
 * HECI HW Section
 */

/* HECI registers */
/* H_CB_WW - Host Circular Buffer (CB) Write Window register */
#define H_CB_WW    0
/* H_CSR - Host Control Status register */
#define H_CSR      4
/* ME_CB_RW - ME Circular Buffer Read Window register (read only) */
#define ME_CB_RW   8
/* ME_CSR_HA - ME Control Status Host Access register (read only) */
#define ME_CSR_HA  0xC


/* register bits of H_CSR (Host Control Status register) */
/* Host Circular Buffer Depth - maximum number of 32-bit entries in CB */
#define H_CBD             0xFF000000
/* Host Circular Buffer Write Pointer */
#define H_CBWP            0x00FF0000
/* Host Circular Buffer Read Pointer */
#define H_CBRP            0x0000FF00
/* Host Reset */
#define H_RST             0x00000010
/* Host Ready */
#define H_RDY             0x00000008
/* Host Interrupt Generate */
#define H_IG              0x00000004
/* Host Interrupt Status */
#define H_IS              0x00000002
/* Host Interrupt Enable */
#define H_IE              0x00000001


/* register bits of ME_CSR_HA (ME Control Status Host Access register) */
/* ME CB (Circular Buffer) Depth HRA (Host Read Access)
 *  - host read only access to ME_CBD */
#define ME_CBD_HRA        0xFF000000
/* ME CB Write Pointer HRA - host read only access to ME_CBWP */
#define ME_CBWP_HRA       0x00FF0000
/* ME CB Read Pointer HRA - host read only access to ME_CBRP */
#define ME_CBRP_HRA       0x0000FF00
/* ME Reset HRA - host read only access to ME_RST */
#define ME_RST_HRA        0x00000010
/* ME Ready HRA - host read only access to ME_RDY */
#define ME_RDY_HRA        0x00000008
/* ME Interrupt Generate HRA - host read only access to ME_IG */
#define ME_IG_HRA         0x00000004
/* ME Interrupt Status HRA - host read only access to ME_IS */
#define ME_IS_HRA         0x00000002
/* ME Interrupt Enable HRA - host read only access to ME_IE */
#define ME_IE_HRA         0x00000001

#define HECI_MINORS_BASE	1
#define HECI_MINORS_COUNT	1

#define  HECI_MINOR_NUMBER	1

/*
 * debug kernel print macro define
 */
extern int heci_debug;
extern struct device *heci_dev;

#define DBG(format, arg...) do { \
	if (heci_debug) { \
		if (heci_dev) \
			dev_info(heci_dev, " %s: " format, \
				 __func__, ## arg); \
		else \
			printk(KERN_INFO "heci: %s: " format, \
				 __func__, ## arg); \
	} \
} while (0)


/*
 * time to wait HECI become ready after init
 */
#define HECI_INTEROP_TIMEOUT    (HZ * 7)

/*
 * watch dog definition
 */
#define HECI_WATCHDOG_DATA_SIZE         16
#define HECI_START_WD_DATA_SIZE         20
#define HECI_WD_PARAMS_SIZE             4
#define HECI_WD_STATE_INDEP             1

/* reserved host client ids (0,1,2) */
#define HECI_WD_HOST_CLIENT_ID          1
#define HECI_IAMTHIF_HOST_CLIENT_ID     2
#define HECI_LAST_FIX_HOST_CLIENT_ID    HECI_IAMTHIF_HOST_CLIENT_ID
/* 256 - 3 reserved */
#define HECI_MAX_OPEN_HANDLE_COUNT      253

#define HECI_EXT_MSG_BUF_LEN            8

struct guid {
	__u32 data1;
	__u16 data2;
	__u16 data3;
	__u8 data4[8];
};

/* File state */
enum file_state {
	HECI_FILE_INITIALIZING = 0,
	HECI_FILE_CONNECTING,
	HECI_FILE_CONNECTED,
	HECI_FILE_DISCONNECTING,
	HECI_FILE_DISCONNECTED
};

/* HECI device states */
enum heci_states {
	HECI_INITIALIZING = 0,
	HECI_ENABLED,
	HECI_RESETING,
	HECI_DISABLED,
	HECI_RECOVERING_FROM_RESET,
	HECI_POWER_DOWN,
	HECI_POWER_UP,
	HECI_REMOVING
};

/* file operation states */
enum heci_file_transaction_states {
	HECI_IDLE = 0,
	HECI_WRITING,
	HECI_WRITE_COMPLETE,
	HECI_FLOW_CONTROL,
	HECI_READING,
	HECI_READ_COMPLETE,
	HECI_CANCELED
};

/* HECI CB */
enum heci_cb_major_types {
	HECI_READ = 0,
	HECI_WRITE,
	HECI_IOCTL,
	HECI_OPEN,
	HECI_CLOSE
};

/* received HECI init message types */
enum heci_received_init_msg_types {
	HECI_MSGINIT_NONE = 0,
	HECI_MSGINIT_INTEROP,
	HECI_MSGINIT_START,
	HECI_MSGINIT_ENUM_CLIENTS,
	HECI_MSGINIT_CLIENT_PROPS
};

/* read completion states */
enum heci_completion_states {
	HECI_WAIT_START = 0,
	HECI_IN_PROGRESS,
	HECI_COMPLETED
};

/* HECI user data struct */
struct heci_message_data {
	__u32 size;
	char *data;
} __attribute__((packed));


#define HECI_CONNECT_TIMEOUT             15	/* seconds */
#define HECI_STALL_TIMER                 12	/* seconds */
#define IAMTHIF_READ_TIMER               15	/* seconds */


/* heci connection struct */
struct heci_connection {
	struct list_head link;
	atomic_t state;
	spinlock_t file_lock; /* file lock */
	/* ID of client connected */
	__u8 host_client_id;
	__u8 me_client_id;
	__u8 flow_ctrl_creds;
	__u8 canceled;
	int status;
	enum heci_file_transaction_states reading_state;
	enum heci_file_transaction_states wr_state;
	struct file *file;
	struct heci_cb_private *read_cb;
	__u32 stall_timer;
	__u32 max_msg_length;
	atomic_t refcnt;
	wait_queue_head_t cc_wait; /* wait for connect or close */
};

struct heci_file_priv {
	struct list_head link;
	struct file *file;
	struct heci_connection *conn;
	wait_queue_head_t tx_wait;
	wait_queue_head_t rx_wait;
	atomic_t sm_state;
	int status;
	atomic_t r_state;
	atomic_t w_state;
	__u8 host_client_id;
};

struct heci_cb_private {
	struct list_head cb_list;
	enum heci_cb_major_types major_file_operations;
	struct heci_connection *conn;
	struct file *file_object;
	struct heci_file_priv *file_ext;
	struct heci_message_data request_buffer;
	struct heci_message_data response_buffer;
	unsigned long information;
	unsigned long read_time;
	atomic_t refcnt;
};

struct heci_io_list {
	struct list_head cb_list;
};

struct heci_driver_version {
	__u8 major;
	__u8 minor;
	__u8 hotfix;
	__u16 build;
} __attribute__((packed));


struct heci_client {
	__u32 max_msg_length;
	__u8 protocol_version;
} __attribute__((packed));

/*
 *  HECI BUS Interface Section
 */
struct heci_msg_hdr {
	__u32 me_addr:8;
	__u32 host_addr:8;
	__u32 length:9;
	__u32 reserved:6;
	__u32 msg_complete:1;
} __attribute__((packed));


struct hbm_cmd {
	__u8 cmd:7;
	__u8 is_response:1;
} __attribute__((packed));


struct heci_bus_message {
	struct hbm_cmd cmd;
	__u8 command_specific_data[];
} __attribute__((packed));

struct hbm_version {
	__u8 minor_version;
	__u8 major_version;
} __attribute__((packed));

struct hbm_host_version_request {
	struct hbm_cmd cmd;
	__u8 reserved;
	struct hbm_version host_version;
} __attribute__((packed));

struct hbm_host_version_response {
	struct hbm_cmd cmd;
	int host_version_supported;
	struct hbm_version me_max_version;
} __attribute__((packed));

struct hbm_host_stop_request {
	struct hbm_cmd cmd;
	__u8 reason;
	__u8 reserved[2];
} __attribute__((packed));

struct hbm_host_stop_response {
	struct hbm_cmd cmd;
	__u8 reserved[3];
} __attribute__((packed));

struct hbm_me_stop_request {
	struct hbm_cmd cmd;
	__u8 reason;
	__u8 reserved[2];
} __attribute__((packed));

struct hbm_host_enum_request {
	struct hbm_cmd cmd;
	__u8 reserved[3];
} __attribute__((packed));

struct hbm_host_enum_response {
	struct hbm_cmd cmd;
	__u8 reserved[3];
	__u8 valid_addresses[32];
} __attribute__((packed));

struct heci_client_properties {
	struct guid protocol_name;
	__u8 protocol_version;
	__u8 max_number_of_connections;
	__u8 fixed_address;
	__u8 single_recv_buf;
	__u32 max_msg_length;
} __attribute__((packed));

struct hbm_props_request {
	struct hbm_cmd cmd;
	__u8 address;
	__u8 reserved[2];
} __attribute__((packed));


struct hbm_props_response {
	struct hbm_cmd cmd;
	__u8 address;
	__u8 status;
	__u8 reserved;
	struct heci_client_properties client_properties;
} __attribute__((packed));

struct hbm_client_connect_request {
	struct hbm_cmd cmd;
	__u8 me_addr;
	__u8 host_addr;
	__u8 reserved;
} __attribute__((packed));

struct hbm_client_connect_response {
	struct hbm_cmd cmd;
	__u8 me_addr;
	__u8 host_addr;
	__u8 status;
} __attribute__((packed));

struct hbm_client_disconnect_request {
	struct hbm_cmd cmd;
	__u8 me_addr;
	__u8 host_addr;
	__u8 reserved;
} __attribute__((packed));

struct hbm_flow_control {
	struct hbm_cmd cmd;
	__u8 me_addr;
	__u8 host_addr;
	__u8 reserved[HECI_FC_MESSAGE_RESERVED_LENGTH];
} __attribute__((packed));

struct heci_me_client {
	struct heci_client_properties props;
	__u8 client_id;
	__u8 flow_ctrl_creds;
} __attribute__((packed));

/* private device struct */
struct iamt_heci_device {
	atomic_t refcnt;
	struct pci_dev *pdev;	/* pointer to pci device struct */
	/*
	 * memory of device
	 */
	unsigned int mem_base;
	unsigned int mem_length;
	char *mem_addr;
	/*
	 * heci device state
	 */
	atomic_t heci_state;
	struct hbm_version version;
	/*
	 * hw states of host and fw(ME)
	 */
	__u32 host_hw_state;
	__u32 me_hw_state;
	atomic_t host_buffer_is_empty;
	spinlock_t hw_lock; /* device hw access lock */

	struct heci_me_client *me_clients; /* Note:memory has to be allocated */
	__u8 num_heci_me_clients;
	__u8 heci_me_clients[32];  /* list of existing clients */
	spinlock_t mecl_lock;      /* lock for me_clients */

	long open_handle_count;
	__u8 current_host_client_id;
	__u8 heci_host_clients[32];   /* bit list of existing clients (32*8) */
	struct list_head conn_list;   /* list of heci connections */
	struct list_head file_list;   /* list of connected files */
	spinlock_t conn_lock; /* lock for host_clients, conn_list, file_list */


	struct heci_io_list pthi_cmd_list;	/* PTHI list for cmd waiting */
	struct heci_io_list heci_cmd_list;	/* write waiting cmd list */
	spinlock_t cmdl_lock;  /* lock for *_cmd_list */

	struct heci_io_list ctrl_wr_list;	/* managed write IOCTL list */
	struct heci_io_list ctrl_rd_list;	/* managed read IOCTL list */
	struct heci_io_list read_list;		/* driver read queue */
	spinlock_t crrdl_lock; /* lock for ctrl_*_list and read_list */

	struct heci_io_list read_cmpl_list;     /* read completed list */
	spinlock_t rdcpl_lock; /* lock for read_cmpl_list */

	struct heci_io_list write_list;		/* driver write queue */
	struct heci_io_list cmpl_list;		/* nonread completed list */

	__u8 need_reset;
	__u8 asf_mode;
	spinlock_t device_lock; /* device lock */

	struct heci_connection wd_conn;
	struct timer_list wd_timer;
	wait_queue_head_t wait_stop_wd;
	atomic_t wd_stoped;
	atomic_t wd_bypass;	/* if 1, don't refresh watchdog ME client */
	__u8 wd_pending;
	__u8 wd_stop;
	__u16 wd_timeout;	/* seconds ((wd_data[1] << 8) + wd_data[0]) */
	__u16 wd_due_counter;
	unsigned char wd_data[HECI_START_WD_DATA_SIZE];

	struct heci_connection iamthif_conn;


	struct task_struct *isr_thread; /* isr processing thread */
	wait_queue_head_t wait_isr; /* wait queue for isr */
	atomic_t recvd_isr;       /* isr process request counter */
	atomic_t recvd_wrr;       /* write request counter */

	struct task_struct *reinit_tsk; /* device reinitialization task */
	struct semaphore reinit_sem; /* reinit task semaphore */
	wait_queue_head_t wait_recvd_msg; /* wait queue for messages from FW */
	atomic_t recvd_msg;


	spinlock_t write_lock;    /* device msg write lock */
	__u32 rd_msg_hdr;         /* read message header */
	__u32 extra_write_index;  /* size of extra control messages */
	__u32 wr_msg_buf[128]; /* msg write buffer used for control messages */
	__u32 rd_msg_buf[128]; /* read buffer used for control messages */
	/* write buffer for extra control messages */
	__u32 ext_msg_buf[HECI_EXT_MSG_BUF_LEN];
};

#endif /* _HECI_DATA_STRUCTURES_H_ */
