/*
 * Part of Intel(R) Manageability Engine Interface Linux driver
 *
 * Copyright (c) 2003 - 2008 Intel Corp.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */

#include "heci.h"


/**
 * _heci_confirm_wd_stoped - confirm watchdog stop completion.
 *
 * @dev: Device object for our driver.
 */
static void _heci_confirm_wd_stoped(struct iamt_heci_device *dev)
{
	if (dev->wd_stop == 2) {
		atomic_set(&dev->wd_stoped, 1);
		wake_up_interruptible(&dev->wait_stop_wd);
	}
}

/**
 * _heci_bh_close - process close related operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 *
 * returns  0, OK; otherwise, error.
 */
static int _heci_bh_close(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn)
{
	DBG("try close client %d ME %d.\n", conn->host_client_id,
	    conn->me_client_id);
	/* check if have not completed writing operation */
	if (conn->wr_state == HECI_WRITING)
		return 0;

	conn->status = 0;

	if (slots == NULL) {
		if (atomic_cmpxchg(&dev->host_buffer_is_empty, 1, 0) != 1)
			return -EMESSAGE_NOT_COMPLETE;
	} else {
		if (!heci_have_slots(slots,
				sizeof(struct hbm_client_disconnect_request)))
			return -EMESSAGE_NOT_COMPLETE;
		*slots -= heci_len_to_slots(
				sizeof(struct hbm_client_disconnect_request));
	}

	priv_cb->information = 0;
	atomic_set(&conn->state, HECI_FILE_DISCONNECTING);
	if (!heci_disconnect(dev, conn)) {
		conn->status = -ENODEV;
		list_del_init(&priv_cb->cb_list);
		heci_free_io_cb(priv_cb);

		if (conn == &dev->wd_conn)
			_heci_confirm_wd_stoped(dev);
		return -ENODEV;
	}

	DBG("ok send close client %d.\n", conn->me_client_id);

	conn->stall_timer = HECI_STALL_TIMER;
	list_move_tail(&priv_cb->cb_list, &dev->ctrl_rd_list.cb_list);
	return 0;
}

/**
 * _heci_hb_start_read - process read related operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 * @cmpl_list: complete list.
 *
 * returns 0, OK; otherwise, error.
 */
static int _heci_bh_start_read(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn,
			struct heci_io_list *cmpl_list)
{
	int ret;

	conn->status = 0;

	ret = heci_start_read_flow(dev, conn, priv_cb, slots);
	if ((ret != 0) && (ret != -EMESSAGE_NOT_COMPLETE))
		list_move_tail(&priv_cb->cb_list, &cmpl_list->cb_list);

	return ret;
}


/**
 * _heci_bh_ioctl - process ioctl related operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 *
 * returns  0, OK; otherwise, error.
 */
static int _heci_bh_ioctl(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn)
{
	if (other_client_is_connecting(dev, conn))
		return 0;

	conn->status = 0;

	if (slots == NULL) {
		if (atomic_cmpxchg(&dev->host_buffer_is_empty, 1, 0) != 1)
			return -EMESSAGE_NOT_COMPLETE;
	} else {
		if (!heci_have_slots(slots,
				sizeof(struct hbm_client_connect_request)))
			return -EMESSAGE_NOT_COMPLETE;
		*slots -= heci_len_to_slots(
				sizeof(struct hbm_client_connect_request));
	}

	priv_cb->information = 0;
	atomic_set(&conn->state, HECI_FILE_CONNECTING);
	if (!heci_connect(dev, conn)) {
		conn->status = -ENODEV;
		list_del_init(&priv_cb->cb_list);
		heci_free_io_cb(priv_cb);
		return -ENODEV;
	}

	conn->stall_timer = HECI_STALL_TIMER;
	list_move_tail(&priv_cb->cb_list, &dev->ctrl_rd_list.cb_list);
	return 0;
}

/**
 * _heci_bh_write - process write operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 * @cmpl_list: complete list.
 *
 * returns  0 - write completed; !=0 - not completed or error.
 */
static int _heci_bh_write(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn,
			struct heci_io_list *cmpl_list)
{
	int res = 0;

	if (flow_ctrl_creds(dev, conn) == 0) {
		DBG("No flow control credentials for client %d.\n",
		    conn->host_client_id);
		return 0;
	}

	spin_lock(&conn->file_lock);

	if (priv_cb->information == 0) {
		if (conn->canceled != 0)
			goto flowctl;
		conn->wr_state = HECI_WRITING;
	}

	if (slots == NULL) {
		if (atomic_cmpxchg(&dev->host_buffer_is_empty, 1, 0) != 1) {
			spin_unlock(&conn->file_lock);
			return -EMESSAGE_WILL_SPLIT;
		}
		res = heci_write_buf(dev, priv_cb, conn,
			HECI_CBD_BYTES, HECI_CBD_BYTES, NULL);
	} else {
		res = heci_write_buf(dev, priv_cb, conn,
			(*slots * sizeof(__u32)), HECI_CBD_BYTES, slots);
	}
	if ((res == -EMESSAGE_NOT_COMPLETE) ||
	    (res == -EMESSAGE_WILL_SPLIT)) {
		spin_unlock(&conn->file_lock);
		return res;
	}

	if (res != 0) {
		spin_unlock(&conn->file_lock);
		list_move_tail(&priv_cb->cb_list, &cmpl_list->cb_list);
		return res;
	}

flowctl:
	conn->reading_state = HECI_FLOW_CONTROL;
	if (conn->canceled == 0)
		conn->wr_state = HECI_WRITE_COMPLETE;

	spin_unlock(&conn->file_lock);

	DBG("write complete\n");
	if (priv_cb->major_file_operations == HECI_READ) {
		spin_lock(&dev->crrdl_lock);
		list_move_tail(&priv_cb->cb_list, &dev->ctrl_wr_list.cb_list);
		spin_lock(&conn->file_lock);
		res = _heci_bh_start_read(dev, slots, priv_cb, conn, cmpl_list);
		spin_unlock(&conn->file_lock);
		spin_unlock(&dev->crrdl_lock);
	} else
		list_move_tail(&priv_cb->cb_list, &cmpl_list->cb_list);

	return res;
}

/**
 * _heci_bh_cmd_write - process write waiting operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 * @cmpl_list: complete list.
 *
 * returns  0 - write completed; !=0 - not completed or error.
 */
static int _heci_bh_cmd_write(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn,
			struct heci_io_list *cmpl_list)
{
	spin_lock(&conn->file_lock);
	priv_cb->information = 0;
	conn->status = 0;
	conn->reading_state = HECI_WRITING;
	list_move_tail(&priv_cb->cb_list, &dev->write_list.cb_list);
	spin_unlock(&conn->file_lock);

	return _heci_bh_write(dev, slots, priv_cb, conn, cmpl_list);
}

/**
 * _heci_bh_ctrl - process waiting control operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 * @cmpl_list: complete list.
 *
 * returns  0 - completed; !=0 - not completed or error.
 */
static int _heci_bh_ctrl(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn,
			struct heci_io_list *cmpl_list)
{
	int ret = 0;

	spin_lock(&conn->file_lock);
	switch (priv_cb->major_file_operations) {
	case HECI_CLOSE:
		ret = _heci_bh_close(dev, slots, priv_cb, conn);
		break;
	case HECI_IOCTL:
		ret = _heci_bh_ioctl(dev, slots, priv_cb, conn);
		break;
	case HECI_READ:
		ret = _heci_bh_start_read(dev, slots, priv_cb, conn, cmpl_list);
		break;
	default:
		BUG();
	}
	spin_unlock(&conn->file_lock);

	return ret;
}

/**
 * _heci_bh_cmd - process waiting operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 * @priv_cb: callback block.
 * @conn: connection data of the file object.
 * @cmpl_list: complete list.
 *
 * returns  0 - completed; !=0 - not completed or error.
 */
static int _heci_bh_cmd(struct iamt_heci_device *dev, __s32 *slots,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn,
			struct heci_io_list *cmpl_list)
{
	int ret;

	if (priv_cb->major_file_operations == HECI_WRITE)
		ret = _heci_bh_cmd_write(dev, slots, priv_cb, conn, cmpl_list);
	else {
		spin_lock(&dev->crrdl_lock);
		list_move_tail(&priv_cb->cb_list, &dev->ctrl_wr_list.cb_list);
		ret = _heci_bh_ctrl(dev, slots, priv_cb, conn, cmpl_list);
		spin_unlock(&dev->crrdl_lock);
	}

	return ret;
}

static void iamthif_state_reset(struct iamt_heci_device *dev)
{
	struct heci_connection *conn = &dev->iamthif_conn;

	spin_lock(&conn->file_lock);
	conn->reading_state = HECI_IDLE;
	conn->wr_state = HECI_IDLE;
	conn->file = NULL;
	conn->read_cb = NULL;
	conn->stall_timer = 0;
	spin_unlock(&conn->file_lock);
}

static int iamthif_state_is_reset(struct iamt_heci_device *dev)
{
	int ret = 0;
	struct heci_connection *conn = &dev->iamthif_conn;

	spin_lock(&conn->file_lock);
	ret = ((conn->reading_state == HECI_IDLE) &&
		(conn->wr_state == HECI_IDLE) &&
		(conn->file == NULL) &&
		(conn->read_cb == NULL));
	spin_unlock(&conn->file_lock);

	return ret;
}

/**
 * run_next_iamthif_cmd - send next message to pthi client
 *
 * @dev: Device object for our driver
 * @slots: free slots.
 *
 * returns 0 on success, <0 on failure.
 */
static void run_next_iamthif_cmd(struct iamt_heci_device *dev, __s32 *slots)
{
	struct heci_connection *conn = &dev->iamthif_conn;
	struct heci_cb_private *priv_cb = NULL;
	int ret = 0;

	if (dev == NULL)
		return;

	iamthif_state_reset(dev);

	spin_lock(&dev->cmdl_lock);

	DBG("complete pthi_cmd_list cb.\n");
	priv_cb = heci_find_list_conn_entry(&dev->pthi_cmd_list, conn);
	if (priv_cb == NULL)
		goto end;

	DBG("write data to pthi client.\n");
	spin_lock(&conn->file_lock);
	conn->file = priv_cb->file_object;
	conn->read_cb = priv_cb;
	spin_unlock(&conn->file_lock);
	ret = _heci_bh_cmd_write(dev, slots, priv_cb, conn, &dev->cmpl_list);
	if ((ret == -EMESSAGE_NOT_COMPLETE) ||
	    (ret == -EMESSAGE_WILL_SPLIT))
		goto end;
	if (ret != 0) {
		DBG("pthi write failed ret = %d\n", ret);
		iamthif_state_reset(dev);
	}
end:
	spin_unlock(&dev->cmdl_lock);
}

/**
 * _heci_cmpl - process completed operation.
 *
 * @dev: Device object for our driver.
 * @priv_cb: callback block.
 * @slots: free slots
 */
static void _heci_cmpl(struct iamt_heci_device *dev,
			struct heci_cb_private *priv_cb, __s32 *slots)
{
	struct heci_connection *conn = priv_cb->conn;
	struct heci_file_priv *file_ext = priv_cb->file_ext;
	enum heci_cb_major_types oper = priv_cb->major_file_operations;

	BUG_ON(conn == NULL);

	if (oper == HECI_READ) {
		spin_lock(&conn->file_lock);
		conn->read_cb = NULL;
		conn->stall_timer = 0;
		spin_unlock(&conn->file_lock);
	}
	if (conn->reading_state == HECI_READ_COMPLETE) {
		spin_lock(&dev->rdcpl_lock);
		list_add_tail(&priv_cb->cb_list, &dev->read_cmpl_list.cb_list);
		spin_unlock(&dev->rdcpl_lock);
		DBG("read completed.\n");
	} else {
		/* cleanup write operations and errors */
		heci_free_io_cb(priv_cb);
		priv_cb = NULL;
	}

	spin_lock(&conn->file_lock);

	if ((file_ext == NULL) || (conn->canceled != 0))
		goto cbfini;

	file_ext->status = conn->status;
	if (oper == HECI_READ) {
		if (conn->reading_state == HECI_READ_COMPLETE)
			atomic_set(&file_ext->r_state, HECI_READ_COMPLETE);
		else
			atomic_set(&file_ext->r_state, HECI_CANCELED);
		if (conn == &dev->iamthif_conn)
			atomic_set(&file_ext->w_state, HECI_IDLE);
		DBG("completing read call back.\n");
		if (waitqueue_active(&file_ext->rx_wait))
			wake_up_interruptible(&file_ext->rx_wait);
	} else if (oper == HECI_WRITE) {
		if (conn->wr_state == HECI_WRITE_COMPLETE)
			atomic_set(&file_ext->w_state, HECI_WRITE_COMPLETE);
		else
			atomic_set(&file_ext->w_state, HECI_CANCELED);
		DBG("completing write call back.\n");
		if (waitqueue_active(&file_ext->tx_wait))
			wake_up_interruptible(&file_ext->tx_wait);
	}

cbfini:
	conn->reading_state = HECI_IDLE;
	conn->wr_state = HECI_IDLE;
	spin_unlock(&conn->file_lock);

	if ((oper == HECI_READ) &&
	    (conn == &dev->iamthif_conn))
		run_next_iamthif_cmd(dev, slots);
}

/**
 * heci_bh_check_ready - check if we are ready to process communication
 *  with heci device, or have to establish communication first.
 *
 * @dev: Device object for our driver
 *
 * returns 0 if we are ready, !=0 if not ready
 */
static int heci_bh_check_ready(struct iamt_heci_device *dev)
{
	struct task_struct *tsk;
	enum heci_states heci_state = atomic_read(&dev->heci_state);

	spin_lock(&dev->hw_lock);
	dev->host_hw_state = read_heci_register(dev, H_CSR);
	dev->me_hw_state = read_heci_register(dev, ME_CSR_HA);

	/* check if ME wants a reset */
	if (((dev->me_hw_state & ME_RDY_HRA) == 0)
	    && (heci_state != HECI_RESETING)
	    && (heci_state != HECI_INITIALIZING)) {
		spin_unlock(&dev->hw_lock);
		DBG("FW not ready.\n");
		heci_reset(dev, 1);
		return 1;
	}

	/* check if we device is started */
	if ((dev->host_hw_state & H_RDY) == H_RDY) {
		spin_unlock(&dev->hw_lock);
		DBG("device is ready.\n");
		return 0;
	}

	/* check if FW is started */
	if ((dev->me_hw_state & ME_RDY_HRA) != ME_RDY_HRA) {
		DBG("enable interrupt FW not ready.\n");
		heci_csr_enable_interrupts(dev);
		spin_unlock(&dev->hw_lock);
		return 2;
	}

	DBG("we need to start the dev.\n");
	dev->host_hw_state |= (H_IE | H_IG | H_RDY);
	heci_set_csr_register(dev);
	spin_unlock(&dev->hw_lock);
	spin_unlock_bh(&dev->device_lock);

	/* check if driver is initializing */
	if (heci_state == HECI_INITIALIZING) {
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_INTEROP);
		wake_up_interruptible(&dev->wait_recvd_msg);
		return 3;
	}

	set_current_state(TASK_INTERRUPTIBLE);
	/* establish communication with heci FW clients */
	tsk = kthread_run(heci_task_initialize_clients,
			  dev, "heci_reinit");
	if (IS_ERR(tsk)) {
		int rc = PTR_ERR(tsk);
		printk(KERN_WARNING "heci: Unable to"
			" start the heci_reinit thread: %d\n", rc);
		return 5;
	}
	return 4;
}

/**
 * _heci_message_good_address - check heci message for valid addresses.
 *
 * @dev: Device object for our driver
 * @heci_hdr: header of heci message
 *
 * returns 0 if ok, <0 if error
 */
static int _heci_message_good_address(struct iamt_heci_device *dev,
				      struct heci_msg_hdr *heci_hdr)
{
	struct heci_connection *conn_pos = NULL;
	struct heci_connection *conn_next = NULL;

	spin_lock(&dev->conn_lock);
	list_for_each_entry_safe(conn_pos, conn_next, &dev->conn_list, link) {
		if ((conn_pos->host_client_id == heci_hdr->host_addr) &&
		    (conn_pos->me_client_id == heci_hdr->me_addr)) {
			spin_unlock(&dev->conn_lock);
			return 0;
		}
	}
	spin_unlock(&dev->conn_lock);
	DBG("corrupted message header\n");
	return -ECORRUPTED_MESSAGE_HEADER;
}

/**
 * _heci_consume_message - discards heci message.
 *
 * @dev: Device object for our driver
 * @heci_hdr: header of heci message
 */
static void _heci_consume_message(struct iamt_heci_device *dev,
				  struct heci_msg_hdr *heci_hdr)
{
	BUG_ON(heci_hdr->length >= sizeof(dev->rd_msg_buf));
	heci_read_slots(dev, (unsigned char *) dev->rd_msg_buf,
				heci_hdr->length);
	DBG("discarding message, header=%08x.\n",
				*(__u32 *) dev->rd_msg_buf);
}

/**
 * heci_same_conn_addr - tell they have same address
 *
 * @conn: connection data of the file object.
 * @resp: (dis)connect response.
 *
 * returns !=0, same; 0,not.
 */
static int heci_same_conn_addr(struct heci_connection *conn,
			     struct hbm_client_connect_response *resp)
{
	return ((conn->host_client_id == resp->host_addr)
		&& (conn->me_client_id == resp->me_addr));
}

/**
 * heci_is_client_connect_msg  - check if the connect message belong
 * to the file connection data.
 *
 * @conn: connection data of the file object
 * @rs: connect response bus message
 *
 * returns 0 on success, <0 on failure.
 */
static int heci_is_client_connect_msg(struct heci_connection *conn,
		struct hbm_client_connect_response *rs)
{
	int ret = 0;

	spin_lock(&conn->file_lock);
	if (heci_same_conn_addr(conn, rs)) {
		if (rs->status == 0) {
			DBG("client connect status = 0x%08x.\n", rs->status);
			atomic_set(&conn->state, HECI_FILE_CONNECTED);
			conn->status = 0;
		} else {
			DBG("client connect status = 0x%08x.\n", rs->status);
			atomic_set(&conn->state, HECI_FILE_DISCONNECTED);
			conn->status = -ENODEV;
		}
		conn->stall_timer = 0;
		ret = 1;
		if (waitqueue_active(&conn->cc_wait))
			wake_up_interruptible(&conn->cc_wait);
	}
	DBG("client state = %d.\n", atomic_read(&conn->state));
	spin_unlock(&conn->file_lock);
	return ret;
}

/**
 * heci_client_connect_response  - connect response bh routine
 *
 * @dev: Device object for our driver
 * @rs: connect response bus message
 *
 * returns 0 on success, <0 on failure.
 */
static int heci_client_connect_response(struct iamt_heci_device *dev,
		struct hbm_client_connect_response *rs)
{
	struct heci_connection *conn;
	struct heci_cb_private *priv_cb_pos = NULL, *priv_cb_next = NULL;
	int ret = -ENODEV;

	/* if WD or iamthif client treat specially */
	if ((heci_is_client_connect_msg(&(dev->wd_conn), rs)) ||
	    (heci_is_client_connect_msg(&(dev->iamthif_conn), rs)))
		return 0;

	spin_lock(&dev->crrdl_lock);
	if (list_empty(&dev->ctrl_rd_list.cb_list)) {
		spin_unlock(&dev->crrdl_lock);
		return ret;
	}

	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
			&dev->ctrl_rd_list.cb_list, cb_list) {
		conn = priv_cb_pos->conn;
		BUG_ON(conn == NULL);
		if (priv_cb_pos->major_file_operations == HECI_IOCTL) {
			if (heci_is_client_connect_msg(conn, rs)) {
				list_del_init(&priv_cb_pos->cb_list);
				heci_free_io_cb(priv_cb_pos);
				ret = 0;
				break;
			}
		}
	}
	spin_unlock(&dev->crrdl_lock);
	return ret;
}

/**
 * heci_client_disconnect_response  - disconnect response bh routine
 *
 * @dev: Device object for our driver
 * @rs: disconnect response bus message
 */
static void heci_client_disconnect_response(struct iamt_heci_device *dev,
					struct hbm_client_connect_response *rs)
{
	struct heci_connection *conn;
	struct heci_cb_private *priv_cb_pos = NULL, *priv_cb_next = NULL;

	spin_lock(&dev->crrdl_lock);
	if (list_empty(&dev->ctrl_rd_list.cb_list)) {
		spin_unlock(&dev->crrdl_lock);
		return;
	}

	DBG("list_for_each_entry_safe in ctrl_rd_list.\n");
	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
			&dev->ctrl_rd_list.cb_list, cb_list) {
		conn = priv_cb_pos->conn;
		BUG_ON(conn == NULL);
		if (heci_same_conn_addr(conn, rs)) {
			spin_lock(&conn->file_lock);
			if (rs->status == 0) {
				atomic_set(&conn->state,
					   HECI_FILE_DISCONNECTED);
			}
			conn->status = rs->status;
			conn->stall_timer = 0;
			list_del_init(&priv_cb_pos->cb_list);
			spin_unlock(&conn->file_lock);
			if (waitqueue_active(&conn->cc_wait))
				wake_up(&conn->cc_wait);
			if (conn == &dev->wd_conn)
				_heci_confirm_wd_stoped(dev);
			heci_free_io_cb(priv_cb_pos);
			break;
		}
	}
	spin_unlock(&dev->crrdl_lock);
}

/**
 * heci_same_flow_addr - tell they have same address.
 *
 * @conn: connection data of the file object.
 * @flow: flow control.
 *
 * returns  !=0, same; 0,not.
 */
static int heci_same_flow_addr(struct heci_connection *conn,
					struct hbm_flow_control *flow)
{
	return ((conn->host_client_id == flow->host_addr)
		&& (conn->me_client_id == flow->me_addr));
}

/**
 * add_single_flow_creds - add single buffer credentials.
 *
 * @file: connection data ot the file object.
 * @flow: flow control.
 *
 * returns 0 on success, <0 on failure.
 */
static int add_single_flow_creds(struct iamt_heci_device *dev,
				  struct hbm_flow_control *flow)
{
	struct heci_me_client *client;
	int i;
	int ret = -ENODEV;

	spin_lock(&dev->mecl_lock);
	for (i = 0; i < dev->num_heci_me_clients; i++) {
		client = &dev->me_clients[i];
		if (flow->me_addr == client->client_id) {
			BUG_ON(client->props.single_recv_buf == 0);
			client->flow_ctrl_creds++;
			DBG("recv flow ctrl msg ME %d (single).\n"
			    "flow control credentials=%d.\n",
			    flow->me_addr, client->flow_ctrl_creds);
			ret = 0;
		}
	}
	spin_unlock(&dev->mecl_lock);
	return ret;
}

/**
 * heci_client_flow_control_response  - flow control response bh routine
 *
 * @dev: Device object for our driver
 * @flow_control: flow control response bus message
 *
 * returns 0 on success, <0 on failure.
 */
static int heci_client_flow_control_response(struct iamt_heci_device *dev,
		struct hbm_flow_control *flow_control)
{
	struct heci_connection *conn_pos = NULL;
	struct heci_connection *conn_next = NULL;
	int ret = -ENODEV;

	if (flow_control->host_addr == 0) {
		/* single receive buffer */
		ret = add_single_flow_creds(dev, flow_control);
		return ret;
	}

	/* normal connection */
	spin_lock(&dev->conn_lock);
	DBG("flow ctrl msg for host %d ME %d.\n",
	    flow_control->host_addr, flow_control->me_addr);
	list_for_each_entry_safe(conn_pos, conn_next, &dev->conn_list, link) {
		DBG("conn of host client %d ME client %d.\n",
		    conn_pos->host_client_id, conn_pos->me_client_id);
		if (heci_same_flow_addr(conn_pos, flow_control)) {
			DBG("recv ctrl msg for host  %d ME %d.\n",
			    flow_control->host_addr, flow_control->me_addr);
			spin_lock(&conn_pos->file_lock);
			conn_pos->flow_ctrl_creds++;
			DBG("flow control credentials=%d.\n",
			    conn_pos->flow_ctrl_creds);
			spin_unlock(&conn_pos->file_lock);
			ret = 0;
			break;
		}
	}
	spin_unlock(&dev->conn_lock);
	return ret;
}

/**
 * heci_same_disconn_addr - tell they have same address
 *
 * @conn: connection data of the file object.
 * @disconn: disconnection request.
 *
 * returns !=0, same; 0,not.
 */
static int heci_same_disconn_addr(struct heci_connection *conn,
			     struct hbm_client_disconnect_request *disconn)
{
	return ((conn->host_client_id == disconn->host_addr)
		&& (conn->me_client_id == disconn->me_addr));
}

/**
 * heci_stop_req - prepare stop request in extra message buffer
 *
 * @dev: Device object for our driver
 */
static void heci_stop_req(struct iamt_heci_device *dev)
{
	struct heci_msg_hdr *heci_hdr;
	struct hbm_host_stop_request *h_stop_req;

	heci_hdr = (struct heci_msg_hdr *)
		    &dev->ext_msg_buf[dev->extra_write_index];
	heci_hdr->host_addr = 0;
	heci_hdr->me_addr = 0;
	heci_hdr->length = sizeof(struct hbm_host_stop_request);
	heci_hdr->msg_complete = 1;
	heci_hdr->reserved = 0;

	h_stop_req = (struct hbm_host_stop_request *)
		    &dev->ext_msg_buf[dev->extra_write_index + 1];
	memset(h_stop_req, 0, sizeof(struct hbm_host_stop_request));
	h_stop_req->cmd.cmd = HOST_STOP_REQ_CMD;
	h_stop_req->reason = DRIVER_STOP_REQUEST;
	dev->extra_write_index += 2;
}

/**
 * heci_disconnect_req - prepare disconnect request in extra message buffer
 *
 * @dev: Device object for our driver
 * @host_client_id: host client id for request
 * @host_client_id: me client id for request
 */
static void heci_disconnect_req(struct iamt_heci_device *dev,
		__u8 host_client_id, __u8 me_client_id)
{
	struct heci_msg_hdr *heci_hdr;
	struct hbm_client_disconnect_request *heci_cli_disconnect;

	heci_hdr = (struct heci_msg_hdr *)
		    &dev->ext_msg_buf[dev->extra_write_index];
	heci_hdr->host_addr = 0;
	heci_hdr->me_addr = 0;
	heci_hdr->length = sizeof(struct hbm_client_disconnect_request);
	heci_hdr->msg_complete = 1;
	heci_hdr->reserved = 0;

	heci_cli_disconnect = (struct hbm_client_disconnect_request *)
		    &dev->ext_msg_buf[dev->extra_write_index + 1];
	memset(heci_cli_disconnect, 0,
		sizeof(struct hbm_client_disconnect_request));
	heci_cli_disconnect->host_addr = host_client_id;
	heci_cli_disconnect->me_addr = me_client_id;
	heci_cli_disconnect->cmd.cmd = CLIENT_DISCONNECT_REQ_CMD;
	dev->extra_write_index += 2;
}

/**
 * heci_disconnect_resp - prepare disconnect response in extra message buffer
 *
 * @dev: Device object for our driver
 * @host_client_id: host client id for response
 * @host_client_id: me client id for response
 */
static void heci_disconnect_resp(struct iamt_heci_device *dev,
		__u8 host_client_id, __u8 me_client_id)
{
	struct heci_msg_hdr *heci_hdr;
	struct hbm_client_connect_response *disconnect_res;

	heci_hdr = (struct heci_msg_hdr *)
		    &dev->ext_msg_buf[dev->extra_write_index];
	heci_hdr->host_addr = 0;
	heci_hdr->me_addr = 0;
	heci_hdr->length = sizeof(struct hbm_client_connect_response);
	heci_hdr->msg_complete = 1;
	heci_hdr->reserved = 0;

	disconnect_res = (struct hbm_client_connect_response *)
			&dev->ext_msg_buf[dev->extra_write_index + 1];
	disconnect_res->host_addr = host_client_id;
	disconnect_res->me_addr = me_client_id;
	*(__u8 *) (&disconnect_res->cmd) = CLIENT_DISCONNECT_RES_CMD;
	disconnect_res->status = 0;
	dev->extra_write_index += 2;
}

/**
 * heci_client_disconnect_request  - disconnect request bh routine
 *
 * @dev: Device object for our driver.
 * @disconnect_req: disconnect request bus message.
 */
static void heci_client_disconnect_request(struct iamt_heci_device *dev,
		struct hbm_client_disconnect_request *disconnect_req)
{
	struct heci_connection *conn_pos = NULL;
	struct heci_connection *conn_next = NULL;

	spin_lock(&dev->conn_lock);
	list_for_each_entry_safe(conn_pos, conn_next, &dev->conn_list, link) {
		if (heci_same_disconn_addr(conn_pos, disconnect_req)) {
			DBG("disconnect request host client %d ME client %d.\n",
					disconnect_req->host_addr,
					disconnect_req->me_addr);
			spin_lock(&conn_pos->file_lock);
			conn_pos->stall_timer = 0;
			atomic_set(&conn_pos->state, HECI_FILE_DISCONNECTED);
			if (conn_pos == &dev->wd_conn) {
				dev->wd_due_counter = 0;
				dev->wd_pending = 0;
			}
			spin_unlock(&conn_pos->file_lock);

			/* prepare disconnect response */
			heci_disconnect_resp(dev,
					     conn_pos->host_client_id,
					     conn_pos->me_client_id);
			break;
		}
	}
	spin_unlock(&dev->conn_lock);
}


static void heci_set_version(struct iamt_heci_device *dev,
			struct hbm_host_version_response *version_res)
{
	if (version_res->host_version_supported) {
		dev->version.major_version = HBM_MAJOR_VERSION;
		dev->version.minor_version = HBM_MINOR_VERSION;
	} else
		dev->version = version_res->me_max_version;
}

/**
 * heci_bh_read_bus_message - bottom half read routine after ISR to
 * handle the read bus message cmd  processing.
 *
 * @dev: Device object for our driver
 * @heci_hdr: header of bus message
 */
static void heci_bh_read_bus_message(struct iamt_heci_device *dev,
		struct heci_msg_hdr *heci_hdr)
{
	struct heci_bus_message *heci_msg;
	struct hbm_client_connect_response *connect_res;
	struct hbm_flow_control *flow_control;
	int ret = 0;
	unsigned char *buffer = (unsigned char *) dev->rd_msg_buf;

	/* read the message to our buffer */
	BUG_ON(heci_hdr->length >= sizeof(dev->rd_msg_buf));
	heci_read_slots(dev, buffer, heci_hdr->length);
	heci_msg = (struct heci_bus_message *) buffer;

	switch (*(__u8 *) heci_msg) {
	case HOST_START_RES_CMD:
		heci_set_version(dev,
			(struct hbm_host_version_response *) heci_msg);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_START);
		DBG("host start response message received.\n");
		break;

	case CLIENT_CONNECT_RES_CMD:
		connect_res =
			(struct hbm_client_connect_response *) heci_msg;
		ret = heci_client_connect_response(dev, connect_res);
		DBG("client connect response message received.\n");
		if (ret != 0)
			heci_disconnect_req(dev,
					    connect_res->host_addr,
					    connect_res->me_addr);
		break;

	case CLIENT_DISCONNECT_RES_CMD:
		heci_client_disconnect_response(dev,
			(struct hbm_client_connect_response *) heci_msg);
		DBG("client disconnect response message received.\n");
		break;

	case HECI_FLOW_CONTROL_CMD:
		flow_control = (struct hbm_flow_control *) heci_msg;
		ret = heci_client_flow_control_response(dev, flow_control);
		DBG("client flow control response message received.\n");
		if (ret != 0)
			heci_disconnect_req(dev,
					    flow_control->host_addr,
					    flow_control->me_addr);
		break;

	case HOST_CLIENT_PROPERTEIS_RES_CMD:
		heci_client_set_props(dev,
				(struct hbm_props_response *) heci_msg);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_CLIENT_PROPS);
		break;

	case HOST_ENUM_RES_CMD:
		spin_lock(&dev->mecl_lock);
		memcpy(dev->heci_me_clients,
			((struct hbm_host_enum_response *)
			    heci_msg)->valid_addresses, 32);
		spin_unlock(&dev->mecl_lock);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_ENUM_CLIENTS);
		break;

	case HOST_STOP_RES_CMD:
		atomic_set(&dev->heci_state, HECI_DISABLED);
		DBG("reseting because of FW stop response.\n");
		heci_reset(dev, 1);
		break;

	case CLIENT_DISCONNECT_REQ_CMD:
		heci_client_disconnect_request(dev,
			(struct hbm_client_disconnect_request *) heci_msg);
		break;

	case ME_STOP_REQ_CMD:
		heci_stop_req(dev);
		break;

	default:
		BUG();
		break;

	}
}

/**
 * _heci_bh_state_ok - check if heci header matches file connection data
 *
 * @conn: connection data of the file object
 * @heci_hdr: header of heci client message
 *
 * returns  !=0 if matches, 0 if no match.
 */
static int _heci_bh_state_ok(struct heci_connection *conn,
					struct heci_msg_hdr *heci_hdr)
{
	return ((conn->host_client_id == heci_hdr->host_addr)
		&& (conn->me_client_id == heci_hdr->me_addr)
		&& (atomic_read(&conn->state) == HECI_FILE_CONNECTED)
		&& (conn->reading_state == HECI_READING)
		&& (conn->read_cb != NULL));
}

/**
 * _heci_bh_read_client_msg - bottom half read routine after ISR to
 * handle the read message data processing.
 *
 * @cmpl_list: An instance of our list structure
 * @dev: Device object for our driver
 * @heci_hdr: header of pthi message
 * @priv_cb: callback block
 * @conn: connection data of the file object
 *
 * returns 0 on success, <0 on failure.
 */
static int _heci_file_read_client_msg(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		struct heci_msg_hdr *heci_hdr,
		struct heci_cb_private *priv_cb,
		struct heci_connection *conn)
{
	unsigned char *buffer;

	BUG_ON(priv_cb->conn != conn);
	BUG_ON(conn->read_cb != priv_cb);
	BUG_ON(heci_hdr->me_addr != conn->me_client_id);
	BUG_ON(conn->reading_state != HECI_READING);
	BUG_ON(priv_cb->response_buffer.data == NULL);

	buffer = (unsigned char *) (priv_cb->response_buffer.data +
				priv_cb->information);
	BUG_ON(priv_cb->response_buffer.size <
				heci_hdr->length + priv_cb->information);
	if (priv_cb->response_buffer.size <
				heci_hdr->length + priv_cb->information) {
		DBG("message overflow.\n");
		list_del_init(&priv_cb->cb_list);
		conn->status = -ENOMEM;
		conn->stall_timer = 0;
		list_add_tail(&priv_cb->cb_list, &cmpl_list->cb_list);
		return -ENOMEM;
	}

	heci_read_slots(dev, buffer, heci_hdr->length);
	priv_cb->information += heci_hdr->length;

	if (heci_hdr->msg_complete == 0)
		return 0;

	list_del_init(&priv_cb->cb_list);
	DBG("completed read host client = %d, ME client = %d, "
	    "data length = %lu\n",
		conn->host_client_id, conn->me_client_id,
		priv_cb->information);

	priv_cb->read_time = get_seconds();
	conn->status = 0;
	conn->stall_timer = 0;
	if (conn->canceled == 0)
		conn->reading_state = HECI_READ_COMPLETE;

	list_add_tail(&priv_cb->cb_list, &cmpl_list->cb_list);
	return 0;
}

/**
 * heci_bh_read_client_message - bottom half read routine after ISR to
 * handle the read heci client message data processing.
 *
 * @cmpl_list: An instance of our list structure
 * @dev: Device object for our driver
 * @heci_hdr: header of heci client message
 *
 * returns  0 on success, <0 on failure.
 */
static int heci_bh_read_client_message(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		struct heci_msg_hdr *heci_hdr)
{
	struct heci_connection *conn;
	struct heci_cb_private *priv_cb_pos = NULL, *priv_cb_next = NULL;
	int ret = -ENODEV;

	DBG("start client msg\n");
	spin_lock(&dev->crrdl_lock);
	if (list_empty(&dev->read_list.cb_list))
		goto quit;

	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
			&dev->read_list.cb_list, cb_list) {
		conn = priv_cb_pos->conn;
		BUG_ON(conn == NULL);
		spin_lock(&conn->file_lock);
		if (!_heci_bh_state_ok(conn, heci_hdr)) {
			spin_unlock(&conn->file_lock);
			continue;
		}
		ret = _heci_file_read_client_msg(cmpl_list, dev,
				heci_hdr, priv_cb_pos, conn);
		spin_unlock(&conn->file_lock);
		break;
	}

quit:
	spin_unlock(&dev->crrdl_lock);
	if (ret != 0)
		_heci_consume_message(dev, heci_hdr);
	return 0;
}

/**
 * heci_bh_read_handler - bottom half read routine after ISR to
 * handle the read processing.
 *
 * @cmpl_list: An instance of our list structure
 * @dev: Device object for our driver
 * @slots: slots to read.
 *
 * returns 0 on success, <0 on failure.
 */
static int heci_bh_read_handler(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		__s32 *slots)
{
	struct heci_msg_hdr *heci_hdr;
	int ret = 0;

	if (!dev->rd_msg_hdr) {
		dev->rd_msg_hdr = read_heci_register(dev, ME_CB_RW);
		DBG("slots=%08x.\n", *slots);
		(*slots)--;
		DBG("slots=%08x.\n", *slots);
	}
	heci_hdr = (struct heci_msg_hdr *) &dev->rd_msg_hdr;
	DBG("heci_hdr->length =%d\n", heci_hdr->length);

	if ((heci_hdr->reserved) || !(dev->rd_msg_hdr)) {
		DBG("corrupted message header.\n");
		ret = -ECORRUPTED_MESSAGE_HEADER;
		goto end;
	}

	if (((*slots) * sizeof(__u32)) < heci_hdr->length) {
		DBG("we can't read the message slots=%08x.\n", *slots);
		/* we can't read the message */
		ret = -ERANGE;
		goto end;
	}

	if ((heci_hdr->host_addr) || (heci_hdr->me_addr)) {
		ret = _heci_message_good_address(dev, heci_hdr);
		if (ret != 0) {
			_heci_consume_message(dev, heci_hdr);
			goto end;
		}
	}
	/* decide where to read the message too */
	if (heci_hdr->host_addr == 0) {
		DBG("call heci_bh_read_bus_message.\n");
		heci_bh_read_bus_message(dev, heci_hdr);
		DBG("end heci_bh_read_bus_message.\n");
	} else {
		DBG("call heci_bh_read_client_message.\n");
		ret = heci_bh_read_client_message(cmpl_list, dev, heci_hdr);
		if (ret != 0)
			goto end;
	}

	/* reset the number of slots and header */
	*slots = count_full_read_slots(dev);
	dev->rd_msg_hdr = 0;

	if (*slots == -ESLOTS_OVERFLOW) {
		/* overflow - reset */
		DBG("reseting due to slots overflow.\n");
		/* set the event since message has been read */
		ret = -ERANGE;
		goto end;
	}
end:
	return ret;
}


/**
 * _heci_bh_ctrl_wr_list_complete - process operations from ctrl_wr_list
 *
 * @cmpl_list: complete list.
 * @dev: Device object for our driver.
 * @slots: free slots.
 *
 * returns  0, OK; otherwise, error.
 */
static int _heci_bh_ctrl_wr_list_complete(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		__s32 *slots)
{
	struct heci_connection *conn;
	struct heci_cb_private *priv_cb_pos = NULL, *priv_cb_next = NULL;
	int ret;

	/* complete control write list CB */
	DBG("complete control write list cb.\n");
	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
				&dev->ctrl_wr_list.cb_list, cb_list) {
		conn = priv_cb_pos->conn;
		BUG_ON(conn == NULL);
		ret = _heci_bh_ctrl(dev, slots, priv_cb_pos, conn, cmpl_list);
		if (ret != 0)
			return ret;
	}
	return 0;
}

/**
 * _heci_bh_write_list_complete - process operations from write_list
 *
 * @cmpl_list: complete list.
 * @dev: Device object for our driver.
 * @slots: free slots.
 *
 * returns  0, OK; otherwise, error.
 */
static int _heci_bh_write_list_complete(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		__s32 *slots)
{
	struct heci_connection *conn;
	struct heci_cb_private *priv_cb_pos = NULL, *priv_cb_next = NULL;
	int ret = 0;

	DBG("complete write list cb.\n");
	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
			&dev->write_list.cb_list, cb_list) {
		conn = priv_cb_pos->conn;
		BUG_ON(conn == NULL);
		ret = _heci_bh_write(dev, slots, priv_cb_pos, conn, cmpl_list);
		if (ret != 0)
			return ret;
	}
	return ret;
}

/**
 * _heci_bh_cmd_list_complete - process operations from heci_cmd_list
 *
 * @cmpl_list: complete list.
 * @dev: Device object for our driver.
 * @slots: free slots.
 *
 * returns  0, OK; otherwise, error.
 */
static int _heci_bh_cmd_list_complete(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		__s32 *slots)
{
	struct heci_connection *conn;
	struct heci_cb_private *priv_cb_pos = NULL, *priv_cb_next = NULL;
	int ret = 0;

	DBG("complete cmd list cb.\n");
	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
			&dev->heci_cmd_list.cb_list, cb_list) {
		conn = priv_cb_pos->conn;
		BUG_ON(conn == NULL);
		ret = _heci_bh_cmd(dev, slots, priv_cb_pos, conn, cmpl_list);
		DBG("cmd ret=%d.\n", ret);
		if (ret != 0)
			return ret;
	}
	return ret;
}

/**
 * heci_bh_write_handler - bottom half write routine after
 * ISR to handle the write processing.
 *
 * @cmpl_list: An instance of our list structure
 * @dev: Device object for our driver
 * @slots: slots to write.
 *
 * returns 0 on success, <0 on failure.
 */
static int heci_bh_write_handler(struct heci_io_list *cmpl_list,
		struct iamt_heci_device *dev,
		__s32 *slots)
{
	int ret;

	host_buffer_is_empty(dev, slots);
	while ((dev->extra_write_index > 0) && (*slots >= 2)) {
		DBG("extra_write_index =%d.\n",	dev->extra_write_index);
		heci_write_message(dev,
				(struct heci_msg_hdr *) &dev->ext_msg_buf[0],
				(unsigned char *) &dev->ext_msg_buf[1],
				sizeof(__u32));
		*slots -= 2;
		dev->extra_write_index -= 2;
		if (dev->extra_write_index > 0)
			memmove(&dev->ext_msg_buf[0], &dev->ext_msg_buf[2],
				dev->extra_write_index);
	}

	if (*slots < 8) {
		DBG("host buffer is full.\n");
		return 0;
	}

	if (atomic_read(&dev->heci_state) == HECI_ENABLED) {
		spin_lock(&dev->wd_conn.file_lock);
		if (dev->wd_pending == 1)
			heci_wd(dev, slots);
		if ((dev->wd_stop == 1) && (dev->wd_pending == 0))
			dev->wd_stop = 2;
		spin_unlock(&dev->wd_conn.file_lock);
	}

	if (atomic_read(&dev->wd_stoped) == 1)
		return -ENODEV;

	/* complete control write list CB */
	spin_lock(&dev->crrdl_lock);
	if (!list_empty(&dev->ctrl_wr_list.cb_list)) {
		ret = _heci_bh_ctrl_wr_list_complete(cmpl_list, dev, slots);
		if (ret != 0) {
			spin_unlock(&dev->crrdl_lock);
			return ret;
		}
	}
	spin_unlock(&dev->crrdl_lock);

	/* complete write list CB */
	if (!list_empty(&dev->write_list.cb_list)) {
		ret = _heci_bh_write_list_complete(cmpl_list, dev, slots);
		if (ret != 0)
			return ret;
	}

	/* call again to complete control write list CB */
	spin_lock(&dev->crrdl_lock);
	if (!list_empty(&dev->ctrl_wr_list.cb_list)) {
		ret = _heci_bh_ctrl_wr_list_complete(cmpl_list, dev, slots);
		if (ret != 0) {
			spin_unlock(&dev->crrdl_lock);
			return ret;
		}
	}
	spin_unlock(&dev->crrdl_lock);

	/* complete cmd list CB */
	if (iamthif_state_is_reset(dev))
		run_next_iamthif_cmd(dev, slots);
	spin_lock(&dev->cmdl_lock);
	atomic_set(&dev->recvd_wrr, 0);
	if (!list_empty(&dev->heci_cmd_list.cb_list)) {
		ret = _heci_bh_cmd_list_complete(cmpl_list, dev, slots);
		if (ret != 0) {
			spin_unlock(&dev->cmdl_lock);
			return ret;
		}
	}
	spin_unlock(&dev->cmdl_lock);

	return 0;
}

/**
 * heci_bh_handler - function called after ISR to handle the interrupt
 *   processing.
 *
 * @dev: Device object for our driver
 * @have_is: call reason - 1 - isr, 2 - user
 */
static void heci_bh_handler(struct iamt_heci_device *dev, int have_is)
{
	__s32 slots;
	int rets;
	struct heci_cb_private *cb_pos = NULL, *cb_next = NULL;

	DBG("function called after ISR to handle the interrupt processing.\n");
	atomic_set(&dev->recvd_wrr, 0);

	if (atomic_read(&dev->heci_state) == HECI_REMOVING)
		return;

	spin_lock_bh(&dev->device_lock);
	if (have_is != 1)
		goto write;
	rets = heci_bh_check_ready(dev);
	if (rets != 0) {
		if (rets < 3)
			spin_unlock_bh(&dev->device_lock);
		return;
	}

	/* check slots avalable for reading */
	slots = count_full_read_slots(dev);
	DBG("slots =%08x  extra_write_index =%08x.\n",
		slots, dev->extra_write_index);
	while ((slots > 0) &&
		(dev->extra_write_index < HECI_EXT_MSG_BUF_LEN)) {
		DBG("slots =%08x  extra_write_index =%08x.\n", slots,
			dev->extra_write_index);
		DBG("call heci_bh_read_handler.\n");
		rets = heci_bh_read_handler(&dev->cmpl_list, dev, &slots);
		if (rets != 0)
			goto end;
	}
write:
	rets = heci_bh_write_handler(&dev->cmpl_list, dev, &slots);
end:
	DBG("end of bottom half function.\n");
	host_buffer_is_empty(dev, &slots);

	if ((atomic_read(&dev->recvd_msg) != HECI_MSGINIT_NONE)
	    && waitqueue_active(&dev->wait_recvd_msg)) {
		DBG("received waiting bus message\n");
		wake_up_interruptible(&dev->wait_recvd_msg);
	}

	if (list_empty(&dev->cmpl_list.cb_list)) {
		host_buffer_is_empty(dev, NULL);
		spin_unlock_bh(&dev->device_lock);
		return;
	}
	list_for_each_entry_safe(cb_pos, cb_next, &dev->cmpl_list.cb_list,
			cb_list) {
		list_del_init(&cb_pos->cb_list);
		_heci_cmpl(dev, cb_pos, &slots);
		cb_pos = NULL;
	}

	host_buffer_is_empty(dev, NULL);
	spin_unlock_bh(&dev->device_lock);
}


/**
 * heci_timer - timer function.
 *
 * @data: pointer to the device structure
 *
 * NOTE: This function is called by timer interrupt work
 */
void heci_wd_timer(unsigned long data)
{
	struct iamt_heci_device *dev = (struct iamt_heci_device *) data;

	DBG("send watchdog.\n");
	spin_lock_bh(&dev->device_lock);

	if ((atomic_read(&dev->heci_state) != HECI_ENABLED)
	    || (atomic_read(&dev->wd_conn.state) != HECI_FILE_CONNECTED))
		goto wdend;

	/* Watchdog */
	if (atomic_read(&dev->wd_bypass) == 0) {
		spin_lock(&dev->wd_conn.file_lock);
		if (dev->wd_due_counter != 0) {
			if (--dev->wd_due_counter == 0) {
				if (heci_wd(dev, NULL) != 0) {
					atomic_inc(&dev->recvd_wrr);
					wake_up_interruptible(&dev->wait_isr);
				}
			}
		}
		spin_unlock(&dev->wd_conn.file_lock);
	}

	spin_lock(&dev->iamthif_conn.file_lock);
	if (dev->iamthif_conn.stall_timer != 0) {
		DBG("stall %d WD\n", dev->iamthif_conn.stall_timer);
		if (--dev->iamthif_conn.stall_timer == 0) {
			spin_unlock(&dev->iamthif_conn.file_lock);
			DBG("reseting because of hang to PTHI.\n");
			heci_reset(dev, 1);
			run_next_iamthif_cmd(dev, NULL);
			spin_lock(&dev->iamthif_conn.file_lock);
		}
	}
	spin_unlock(&dev->iamthif_conn.file_lock);
wdend:
	mod_timer(&dev->wd_timer, round_jiffies(jiffies + 2 * HZ));
	spin_unlock_bh(&dev->device_lock);
}


/**
 * heci_isr_thread - isr worker thread.
 *
 * @data: pointer to the device structure
 *
 * returns 1 if failed to start, 0 if worked
 */
int heci_isr_thread(void *data)
{
	int have_is;
	int err;
	struct iamt_heci_device *dev = (struct iamt_heci_device *) data;
	struct task_struct *tsk = current;
	struct sched_param param = { .sched_priority = 1 };

	if (dev == NULL)
		return 1;

	get_heci_device(dev);

	sched_setscheduler(tsk, SCHED_FIFO, &param);

	for (;;) {
		err = wait_event_interruptible(dev->wait_isr,
			 ((atomic_read(&dev->recvd_isr) != 0)
			  || (atomic_read(&dev->recvd_wrr) != 0)
			  || kthread_should_stop()
			  || (atomic_read(&dev->heci_state) == HECI_REMOVING)));
		if (err == -ERESTARTSYS)
			goto tend;

		if (kthread_should_stop()
		    || (atomic_read(&dev->heci_state) == HECI_REMOVING))
			goto tend;

		have_is = 0;
		if (atomic_read(&dev->recvd_isr) > 0) {
			atomic_set(&dev->recvd_isr, 0);

			spin_lock_bh(&dev->hw_lock);
			dev->host_hw_state = read_heci_register(dev, H_CSR);
			if ((dev->host_hw_state & H_IS) == H_IS) {
				/* acknowledge interrupt */
				heci_set_csr_register(dev);
				have_is = 1;
			}
			spin_unlock_bh(&dev->hw_lock);

		} else if (atomic_read(&dev->recvd_wrr) > 0)
			have_is = 2;
		if (have_is != 0)
			heci_bh_handler(dev, have_is);
	}

tend:
	spin_lock_bh(&dev->device_lock);
	dev->isr_thread = NULL;
	spin_unlock_bh(&dev->device_lock);

	put_heci_device(dev);

	return 0;
}

/**
 * heci_isr_interrupt - The ISR of the HECI device
 *
 * @irq: The irq number
 * @dev_id: pointer to the device structure
 *
 * returns irqreturn_t
 */
irqreturn_t heci_isr_interrupt(int irq, void *dev_id)
{
	__u32 host_hw_state;
	struct iamt_heci_device *dev = (struct iamt_heci_device *) dev_id;

	host_hw_state = read_heci_register(dev, H_CSR);
	if ((host_hw_state & H_IS) != H_IS)
		return IRQ_NONE;

	DBG("ISR\n");
	atomic_inc(&dev->recvd_isr);
	wake_up_interruptible(&dev->wait_isr);

	return IRQ_HANDLED;
}

